@@ -41,12 +41,10 @@
 #include <errno.h>
 #include <arpa/inet.h>
 #include <netinet/in.h>
-#include <pthread.h>
 #include <sys/un.h>
 #endif
 
 #include <libcork/core.h>
-#include <udns.h>
 
 #ifdef __MINGW32__
 #include "win32.h"
@@ -74,7 +72,7 @@
 #endif
 
 #ifndef BUF_SIZE
-#define BUF_SIZE 2048
+#define BUF_SIZE 16384
 #endif
 
 #ifndef SSMAXCONN
@@ -89,6 +87,7 @@ static void remote_recv_cb(EV_P_ ev_io *w, int revents);
 static void remote_send_cb(EV_P_ ev_io *w, int revents);
 static void server_timeout_cb(EV_P_ ev_timer *watcher, int revents);
 
+static void perform_handshake(EV_P_ server_t *server);
 static remote_t *new_remote(int fd);
 static server_t *new_server(int fd, listen_ctx_t *listener);
 static remote_t *connect_to_remote(EV_P_ struct addrinfo *res,
@@ -98,12 +97,11 @@ static void free_remote(remote_t *remote);
 static void close_and_free_remote(EV_P_ remote_t *remote);
 static void free_server(server_t *server);
 static void close_and_free_server(EV_P_ server_t *server);
-static void server_resolve_cb(struct sockaddr *addr, void *data);
-static void query_free_cb(void *data);
 
 int verbose = 0;
 
 static int ipv6first = 0;
+static int reverse_proxy = 0;
 static int fast_open = 0;
 
 static obfs_para_t *obfs_para = NULL;
@@ -392,6 +390,109 @@ connect_to_remote(EV_P_ struct addrinfo *res,
     return remote;
 }
 
+static void
+perform_handshake(EV_P_ server_t *server)
+{
+    // Copy back the saved first packet
+    server->buf->len = server->header_buf->len;
+    server->buf->idx = server->header_buf->idx;
+    memcpy(server->buf->data, server->header_buf->data, server->header_buf->len);
+    server->header_buf->idx = server->header_buf->len = 0;
+
+    struct addrinfo info;
+    struct sockaddr_storage storage;
+    memset(&info, 0, sizeof(struct addrinfo));
+    memset(&storage, 0, sizeof(struct sockaddr_storage));
+
+    // Domain name
+    size_t name_len = strlen(server->listen_ctx->dst_addr->host);
+    char *host = server->listen_ctx->dst_addr->host;
+    uint16_t port = htons((uint16_t)atoi(server->listen_ctx->dst_addr->port));
+
+    if (obfs_para == NULL || !obfs_para->is_enable(server->obfs)) {
+        if (server->listen_ctx->failover->host != NULL
+                && server->listen_ctx->failover->port != NULL) {
+            name_len = strlen(server->listen_ctx->failover->host);
+            host = server->listen_ctx->failover->host;
+            port = htons((uint16_t)atoi(server->listen_ctx->failover->port));
+        }
+    }
+
+    struct cork_ip ip;
+    if (cork_ip_init(&ip, host) != -1) {
+        if (ip.version == 4) {
+            struct sockaddr_in *addr = (struct sockaddr_in *)&storage;
+            inet_pton(AF_INET, host, &(addr->sin_addr));
+            addr->sin_port   = port;
+            addr->sin_family = AF_INET;
+        } else if (ip.version == 6) {
+            struct sockaddr_in6 *addr = (struct sockaddr_in6 *)&storage;
+            inet_pton(AF_INET6, host, &(addr->sin6_addr));
+            addr->sin6_port   = port;
+            addr->sin6_family = AF_INET6;
+        }
+    } else {
+        if (!validate_hostname(host, name_len)) {
+            LOGE("invalid host name");
+            close_and_free_server(EV_A_ server);
+            return;
+        }
+        char tmp_port[16];
+        snprintf(tmp_port, 16, "%d", ntohs(port));
+        if (get_sockaddr(host, tmp_port, &storage, 0, ipv6first) == -1) {
+            LOGE("failed to resolve the provided hostname");
+            close_and_free_server(EV_A_ server);
+            return;
+        }
+    }
+
+    info.ai_socktype = SOCK_STREAM;
+    info.ai_protocol = IPPROTO_TCP;
+
+    if (storage.ss_family == AF_INET) {
+        info.ai_family   = AF_INET;
+        info.ai_addrlen  = sizeof(struct sockaddr_in);
+        info.ai_addr     = (struct sockaddr *)&storage;
+    } else if (storage.ss_family == AF_INET6) {
+        info.ai_family   = AF_INET6;
+        info.ai_addrlen  = sizeof(struct sockaddr_in6);
+        info.ai_addr     = (struct sockaddr *)&storage;
+    } else {
+        LOGE("failed to resolve the provided hostname");
+        close_and_free_server(EV_A_ server);
+        return;
+    }
+
+    if (verbose) {
+        LOGI("connect to %s:%d", host, ntohs(port));
+    }
+
+    remote_t *remote = connect_to_remote(EV_A_ & info, server);
+
+    if (remote == NULL) {
+        LOGE("connect error");
+        close_and_free_server(EV_A_ server);
+        return;
+    } else {
+        server->remote = remote;
+        remote->server = server;
+
+        // XXX: should handle buffer carefully
+        if (server->buf->len > 0) {
+            memcpy(remote->buf->data, server->buf->data, server->buf->len);
+            remote->buf->len = server->buf->len;
+            remote->buf->idx = 0;
+            server->buf->len = 0;
+            server->buf->idx = 0;
+        }
+
+        // waiting on remote connected event
+        ev_io_start(EV_A_ & remote->send_ctx->io);
+    }
+
+    return;
+}
+
 static void
 server_recv_cb(EV_P_ ev_io *w, int revents)
 {
@@ -461,7 +562,49 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
                 obfs_para->disable(server->obfs);
             }
         }
+
         server->stage = STAGE_HANDSHAKE;
+        ev_io_stop(EV_A_ & server->recv_ctx->io);
+
+        // Copy the first packet to the currently unused header_buf.
+        server->header_buf->len = server->buf->len - server->buf->idx;
+        server->header_buf->idx = 0;
+        memcpy(server->header_buf->data, server->buf->data + server->buf->idx, server->header_buf->len);
+        if (reverse_proxy && obfs_para->send_empty_response_upon_connection) {
+            // Clear the buffer to make an empty packet.
+            server->buf->len = 0;
+
+            if (obfs_para) {
+                obfs_para->obfs_response(server->buf, BUF_SIZE, server->obfs);
+            }
+
+            int s = send(server->fd, server->buf->data, server->buf->len, 0);
+
+            if (s == -1) {
+                if (errno == EAGAIN || errno == EWOULDBLOCK) {
+                    // no data, wait for send
+                    server->buf->idx = 0;
+                    ev_io_start(EV_A_ & server->send_ctx->io);
+                    return;
+                } else {
+                    ERROR("send_inital_response");
+                    close_and_free_remote(EV_A_ remote);
+                    close_and_free_server(EV_A_ server);
+                    return;
+                }
+            } else if (s < server->buf->len) {
+                server->buf->len -= s;
+                server->buf->idx  = s;
+                ev_io_start(EV_A_ & server->send_ctx->io);
+                return;
+            } else {
+                server->buf->len = 0;
+                server->buf->idx = 0;
+            }
+        }
+
+        perform_handshake(EV_A_ server);
+        return;
     } else {
         buf->len = r;
         if (obfs_para) {
@@ -493,100 +636,6 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
         }
         return;
 
-    } else if (server->stage == STAGE_HANDSHAKE) {
-
-        int need_query = 0;
-        struct addrinfo info;
-        struct sockaddr_storage storage;
-        memset(&info, 0, sizeof(struct addrinfo));
-        memset(&storage, 0, sizeof(struct sockaddr_storage));
-
-        // Domain name
-        size_t name_len = strlen(server->listen_ctx->dst_addr->host);
-        char *host = server->listen_ctx->dst_addr->host;
-        uint16_t port = htons((uint16_t)atoi(server->listen_ctx->dst_addr->port));
-
-        if (obfs_para == NULL || !obfs_para->is_enable(server->obfs)) {
-            if (server->listen_ctx->failover->host != NULL
-                    && server->listen_ctx->failover->port != NULL) {
-                name_len = strlen(server->listen_ctx->failover->host);
-                host = server->listen_ctx->failover->host;
-                port = htons((uint16_t)atoi(server->listen_ctx->failover->port));
-            }
-        }
-
-        struct cork_ip ip;
-        if (cork_ip_init(&ip, host) != -1) {
-            info.ai_socktype = SOCK_STREAM;
-            info.ai_protocol = IPPROTO_TCP;
-            if (ip.version == 4) {
-                struct sockaddr_in *addr = (struct sockaddr_in *)&storage;
-                dns_pton(AF_INET, host, &(addr->sin_addr));
-                addr->sin_port   = port;
-                addr->sin_family = AF_INET;
-                info.ai_family   = AF_INET;
-                info.ai_addrlen  = sizeof(struct sockaddr_in);
-                info.ai_addr     = (struct sockaddr *)addr;
-            } else if (ip.version == 6) {
-                struct sockaddr_in6 *addr = (struct sockaddr_in6 *)&storage;
-                dns_pton(AF_INET6, host, &(addr->sin6_addr));
-                addr->sin6_port   = port;
-                addr->sin6_family = AF_INET6;
-                info.ai_family    = AF_INET6;
-                info.ai_addrlen   = sizeof(struct sockaddr_in6);
-                info.ai_addr      = (struct sockaddr *)addr;
-            }
-        } else {
-            if (!validate_hostname(host, name_len)) {
-                LOGE("invalid host name");
-                close_and_free_server(EV_A_ server);
-                return;
-            }
-            need_query = 1;
-        }
-
-        if (verbose) {
-            LOGI("connect to %s:%d", host, ntohs(port));
-        }
-
-        if (!need_query) {
-            remote_t *remote = connect_to_remote(EV_A_ & info, server);
-
-            if (remote == NULL) {
-                LOGE("connect error");
-                close_and_free_server(EV_A_ server);
-                return;
-            } else {
-                server->remote = remote;
-                remote->server = server;
-
-                // XXX: should handle buffer carefully
-                if (server->buf->len > 0) {
-                    memcpy(remote->buf->data, server->buf->data, server->buf->len);
-                    remote->buf->len = server->buf->len;
-                    remote->buf->idx = 0;
-                    server->buf->len = 0;
-                    server->buf->idx = 0;
-                }
-
-                // waiting on remote connected event
-                ev_io_stop(EV_A_ & server_recv_ctx->io);
-                ev_io_start(EV_A_ & remote->send_ctx->io);
-            }
-        } else {
-            query_t *query = ss_malloc(sizeof(query_t));
-            memset(query, 0, sizeof(query_t));
-            query->server = server;
-            snprintf(query->hostname, 256, "%s", host);
-
-            server->stage = STAGE_RESOLVE;
-            server->query = resolv_query(host, server_resolve_cb,
-                                         query_free_cb, query, port);
-
-            ev_io_stop(EV_A_ & server_recv_ctx->io);
-        }
-
-        return;
     }
     // should not reach here
     FATAL("server context error");
@@ -634,14 +683,21 @@ server_send_cb(EV_P_ ev_io *w, int revents)
             server->buf->len = 0;
             server->buf->idx = 0;
             ev_io_stop(EV_A_ & server_send_ctx->io);
-            if (remote != NULL) {
-                ev_io_start(EV_A_ & remote->recv_ctx->io);
-                return;
-            } else {
-                LOGE("invalid remote");
-                close_and_free_remote(EV_A_ remote);
-                close_and_free_server(EV_A_ server);
+
+            // If handshaking
+            if (server->stage == STAGE_HANDSHAKE) {
+                perform_handshake(EV_A_ server);
                 return;
+            } else { // If streaming
+                if (remote != NULL) {
+                    ev_io_start(EV_A_ & remote->recv_ctx->io);
+                    return;
+                } else {
+                    LOGE("invalid remote");
+                    close_and_free_remote(EV_A_ remote);
+                    close_and_free_server(EV_A_ server);
+                    return;
+                }
             }
         }
     }
@@ -663,69 +719,6 @@ server_timeout_cb(EV_P_ ev_timer *watcher, int revents)
     close_and_free_server(EV_A_ server);
 }
 
-static void
-query_free_cb(void *data)
-{
-    if (data != NULL) {
-        ss_free(data);
-    }
-}
-
-static void
-server_resolve_cb(struct sockaddr *addr, void *data)
-{
-    query_t *query       = (query_t *)data;
-    server_t *server     = query->server;
-    struct ev_loop *loop = server->listen_ctx->loop;
-
-    server->query = NULL;
-
-    if (addr == NULL) {
-        LOGE("unable to resolve %s", query->hostname);
-        close_and_free_server(EV_A_ server);
-    } else {
-        if (verbose) {
-            LOGI("successfully resolved %s", query->hostname);
-        }
-
-        struct addrinfo info;
-        memset(&info, 0, sizeof(struct addrinfo));
-        info.ai_socktype = SOCK_STREAM;
-        info.ai_protocol = IPPROTO_TCP;
-        info.ai_addr     = addr;
-
-        if (addr->sa_family == AF_INET) {
-            info.ai_family  = AF_INET;
-            info.ai_addrlen = sizeof(struct sockaddr_in);
-        } else if (addr->sa_family == AF_INET6) {
-            info.ai_family  = AF_INET6;
-            info.ai_addrlen = sizeof(struct sockaddr_in6);
-        }
-
-        remote_t *remote = connect_to_remote(EV_A_ & info, server);
-
-        if (remote == NULL) {
-            close_and_free_server(EV_A_ server);
-        } else {
-            server->remote = remote;
-            remote->server = server;
-
-            // XXX: should handle buffer carefully
-            if (server->buf->len > 0) {
-                memcpy(remote->buf->data, server->buf->data + server->buf->idx,
-                       server->buf->len);
-                remote->buf->len = server->buf->len;
-                remote->buf->idx = 0;
-                server->buf->len = 0;
-                server->buf->idx = 0;
-            }
-
-            // listen to remote connected event
-            ev_io_start(EV_A_ & remote->send_ctx->io);
-        }
-    }
-}
-
 static void
 remote_recv_cb(EV_P_ ev_io *w, int revents)
 {
@@ -973,7 +966,6 @@ new_server(int fd, listen_ctx_t *listener)
     server->send_ctx->server    = server;
     server->send_ctx->connected = 0;
     server->stage               = STAGE_INIT;
-    server->query               = NULL;
     server->listen_ctx          = listener;
     server->remote              = NULL;
 
@@ -1027,10 +1019,6 @@ static void
 close_and_free_server(EV_P_ server_t *server)
 {
     if (server != NULL) {
-        if (server->query != NULL) {
-            resolv_cancel(server->query);
-            server->query = NULL;
-        }
         ev_io_stop(EV_A_ & server->send_ctx->io);
         ev_io_stop(EV_A_ & server->recv_ctx->io);
         ev_timer_stop(EV_A_ & server->recv_ctx->watcher);
@@ -1096,8 +1084,7 @@ main(int argc, char **argv)
     int server_num = 0;
     const char *server_host[MAX_REMOTE_NUM];
 
-    char *nameservers[MAX_DNS_NUM + 1];
-    int nameserver_num = 0;
+    char *nameservers = NULL;
 
     ss_addr_t dst_addr = { .host = NULL, .port = NULL };
     char *dst_addr_str = NULL;
@@ -1180,6 +1167,8 @@ main(int argc, char **argv)
                     obfs_host = value;
                 } else if (strcmp(key, "failover") == 0) {
                     failover_str = value;
+                } else if (strcmp(key, "reverse_proxy") == 0) {
+                    reverse_proxy = 1;
 #ifdef __linux__
                 } else if (strcmp(key, "mptcp") == 0) {
                     mptcp = 1;
@@ -1200,6 +1189,7 @@ main(int argc, char **argv)
 #ifdef __linux__
         { "mptcp",           no_argument,       0, 0 },
 #endif
+        { "reverse_proxy",   no_argument,       0, 0 },
         { 0,                 0,                 0, 0 }
     };
 
@@ -1228,6 +1218,9 @@ main(int argc, char **argv)
             } else if (option_index == 5) {
                 mptcp = 1;
                 LOGI("enable multipath TCP");
+            } else if (option_index == 6) {
+                reverse_proxy = 1;
+                LOGI("enable reverse proxy");
             }
             break;
         case 's':
@@ -1258,9 +1251,7 @@ main(int argc, char **argv)
             iface = optarg;
             break;
         case 'd':
-            if (nameserver_num < MAX_DNS_NUM) {
-                nameservers[nameserver_num++] = optarg;
-            }
+            nameservers = optarg;
             break;
         case 'a':
             user = optarg;
@@ -1336,12 +1327,15 @@ main(int argc, char **argv)
             nofile = conf->nofile;
         }
 #endif
-        if (conf->nameserver != NULL) {
-            nameservers[nameserver_num++] = conf->nameserver;
+        if (nameservers == NULL) {
+            nameservers = conf->nameserver;
         }
         if (ipv6first == 0) {
             ipv6first = conf->ipv6_first;
         }
+        if (reverse_proxy == 0) {
+            reverse_proxy = conf->reverse_proxy;
+        }
     }
 
     if (server_num == 0) {
@@ -1428,20 +1422,8 @@ main(int argc, char **argv)
     // initialize ev loop
     struct ev_loop *loop = EV_DEFAULT;
 
-    // setup udns
-    if (nameserver_num == 0) {
-#ifdef __MINGW32__
-        nameservers[nameserver_num++] = "8.8.8.8";
-        resolv_init(loop, nameservers, nameserver_num, ipv6first);
-#else
-        resolv_init(loop, NULL, 0, ipv6first);
-#endif
-    } else {
-        resolv_init(loop, nameservers, nameserver_num, ipv6first);
-    }
-
-    for (int i = 0; i < nameserver_num; i++)
-        LOGI("using nameserver: %s", nameservers[i]);
+    if (nameservers != NULL)
+        LOGI("using nameserver: %s", nameservers);
 
     // initialize listen context
     listen_ctx_t listen_ctx_list[server_num];
@@ -1516,8 +1498,6 @@ main(int argc, char **argv)
 
     free_connections(loop);
 
-    resolv_shutdown(loop);
-
 #ifdef __MINGW32__
     winsock_cleanup();
 #endif
